//---------------------------------------------------------------------------
// 2E_Jacobi_parallel_submatrices.cpp
//---------------------------------------------------------------------------
#if (defined __linux__) || (defined _AIX) || (defined __APPLE__)
	#include <sys/types.h>
	#include <sys/stat.h>
	#include <unistd.h>
#elif (defined _WIN32) || (defined _WIN64)
	#include <conio.h>
	#include <direct.h>
#endif

#include <mpi.h>
#include <math.h>
#include <time.h>
#include <iostream>
using namespace std;

#define N 20
//---------------------------------------------------------------------------
int main(int argc, char* argv[])
{
	int iterations, i, j, k, rank, size, N_sqrt;
	int row_region, column_region, row_start, row_end, column_start, column_end;
	double my_time, A[N + 2][N + 2], B[N + 2][N + 2];
	MPI_Datatype mytype1, mytype2, mytype3, mytype4;

	int src, dest;
	int dims[2], periods[2];
	MPI_Comm mycomm_cart;

	MPI_Init(&argc, &argv);
	MPI_Barrier(MPI_COMM_WORLD);
	my_time = -MPI_Wtime();

	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	MPI_Comm_size(MPI_COMM_WORLD, &size);
	if (rank == 0) {
		printf("There are %d processes.\n", size);
		fflush(stdout);
	}

	if (!(size == 4 || size == 16 || size == 25 || size == 100)) {
		if (rank == 0)
			printf("The number of processes should be 4, 16, 25, or 100!\n");
		goto end;
	}

	switch (size) {
	case 4: iterations = 10000000; break;
	case 16: iterations = 1000; break;
	case 25: iterations = 1000; break;
	case 100: iterations = 100; break;
	default: break;
	}

	N_sqrt = (int)sqrt(size);
	row_region = N / N_sqrt;
	column_region = N / N_sqrt;
	row_start = (rank % N_sqrt) * row_region + 1;
	row_end = row_start + row_region - 1;
	column_start = (rank / N_sqrt) * column_region + 1;
	column_end = column_start + column_region - 1;

	MPI_Type_vector(row_region, 1, 1, MPI_DOUBLE, &mytype1);
	MPI_Type_commit(&mytype1);
	MPI_Type_vector(column_region, 1, N + 2, MPI_DOUBLE, &mytype2);
	MPI_Type_commit(&mytype2);
	MPI_Type_vector(column_region, row_region, N + 2, MPI_DOUBLE, &mytype3);
	MPI_Type_commit(&mytype3);
	MPI_Type_vector((N + 2) * (N + 2), 1, 1, MPI_DOUBLE, &mytype4);
	MPI_Type_commit(&mytype4);

	dims[0] = size / N_sqrt;
	dims[1] = N_sqrt;
	periods[0] = 1;
	periods[1] = 1;

	MPI_Cart_create(MPI_COMM_WORLD, 2, dims, periods, true, &mycomm_cart);

	printf("Process %d: row_start %d, row_end %d column_start %d, column_end %d.\n", 
		rank, row_start, row_end, column_start, column_end);
	fflush(stdout);

	for (i = 0; i < N + 2; i++)
		for (j = 0; j < N + 2; j++)
			A[i][j] = B[i][j] = 0;

	if (rank == 0) {
		srand((unsigned)time(NULL));
		for (i = 0; i < N + 2; i++)
			for (j = 0; j < N + 2; j++)
				A[i][j] = rand() % 100;
	}

	if (rank == 0)
		for (k = 1; k < size; k++)
			MPI_Ssend(&(A[0][0]), 1, mytype4, k, 0, MPI_COMM_WORLD);
	if (rank > 0)
		MPI_Recv(&(A[0][0]), 1, mytype4, 0, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);

	if (rank == 0) {
		cout << "\nStarting values of a matrix A:\n";
		for (i = 0; i < N + 2; i++) {
			for (j = 0; j < N + 2; j++)
				printf("%6.2f ", A[i][j]);
			cout << endl;
		}
	}
	fflush(stdout);

	for (k = 0; k < iterations; k++) {
		for (i = row_start; i <= row_end; i++)
			for (j = column_start; j <= column_end; j++)
				B[i][j] = 0.25 * (A[i - 1][j] + A[i + 1][j] + A[i][j - 1] + A[i][j + 1]);

		for (i = row_start; i <= row_end; i++)
			for (j = column_start; j <= column_end; j++)
				A[i][j] = B[i][j];

		// posielanie riadkovych halo po vertikalach
		// zhora dole
		MPI_Cart_shift(mycomm_cart, 1, 1, &src, &dest);
		if (rank < dest)
			MPI_Ssend(&(B[row_end][column_start]), 1, mytype1, dest, rank, mycomm_cart);
		if (src < rank)
			MPI_Recv(&(A[row_start - 1][column_start]), 1, mytype1, src, src, 
				mycomm_cart, MPI_STATUS_IGNORE);

		// zdola hore
		MPI_Cart_shift(mycomm_cart, 1, -1, &src, &dest);
		if (rank > dest)
			MPI_Ssend(&(B[row_start][column_start]), 1, mytype1, dest, size + rank, 
				mycomm_cart);
		if (src > rank)
			MPI_Recv(&(A[row_end + 1][column_start]), 1, mytype1, src, size + src, 
				mycomm_cart, MPI_STATUS_IGNORE);

		// posielanie stlpcovych halo po horizontalach
		// zlava doprava
		MPI_Cart_shift(mycomm_cart, 0, 1, &src, &dest);
		if (rank < dest)
			MPI_Ssend(&(B[row_start][column_end]), 1, mytype2, dest, 2 * size + rank, 
				mycomm_cart);
		if (src < rank)
			MPI_Recv(&(A[row_start][column_start - 1]), 1, mytype2, src, 2 * size + src,
				mycomm_cart, MPI_STATUS_IGNORE);

		// zprava dolava
		MPI_Cart_shift(mycomm_cart, 0, -1, &src, &dest);
		if (rank > dest && dest >= 0)
			MPI_Ssend(&(B[row_start][column_start]), 1, mytype2, dest, 3 * size + rank, 
				mycomm_cart);
		if (src > rank)
			MPI_Recv(&(A[row_start][column_end + 1]), 1, mytype2, src, 3 * size + src, 
				mycomm_cart, MPI_STATUS_IGNORE);

		MPI_Barrier(MPI_COMM_WORLD);
	}

	if (rank > 0) {
		MPI_Ssend(&(A[row_start][column_start]), 1, mytype3, 0, 0, MPI_COMM_WORLD);
	}
	if (rank == 0)
		for (k = 1; k < size; k++)
			MPI_Recv(&(A[(k % N_sqrt) * row_region + 1][(k / N_sqrt) * column_region
				+ 1]), 1, mytype3, k, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);

	if (rank == 0) {
		cout << "\nFinal values of the matrix A:\n";
		for (i = 0; i < N + 2; i++) {
			for (j = 0; j < N + 2; j++)
				printf("%6.2f ", A[i][j]);
			cout << endl;
		}
	}

end:
	my_time += MPI_Wtime();
	if (rank == 0)
		printf("\nTime = %10.6f s\n", my_time);

	fflush(stdout);
	MPI_Finalize();

	return 0;
}